lib/zlib: add s390 hardware support for kernel zlib_deflate
[linux-2.6-microblaze.git] / mm / slub.c
index 8eafccf..17dc00e 100644 (file)
--- a/mm/slub.c
+++ b/mm/slub.c
@@ -439,19 +439,38 @@ static inline bool cmpxchg_double_slab(struct kmem_cache *s, struct page *page,
 }
 
 #ifdef CONFIG_SLUB_DEBUG
+static unsigned long object_map[BITS_TO_LONGS(MAX_OBJS_PER_PAGE)];
+static DEFINE_SPINLOCK(object_map_lock);
+
 /*
  * Determine a map of object in use on a page.
  *
  * Node listlock must be held to guarantee that the page does
  * not vanish from under us.
  */
-static void get_map(struct kmem_cache *s, struct page *page, unsigned long *map)
+static unsigned long *get_map(struct kmem_cache *s, struct page *page)
 {
        void *p;
        void *addr = page_address(page);
 
+       VM_BUG_ON(!irqs_disabled());
+
+       spin_lock(&object_map_lock);
+
+       bitmap_zero(object_map, page->objects);
+
        for (p = page->freelist; p; p = get_freepointer(s, p))
-               set_bit(slab_index(p, s, addr), map);
+               set_bit(slab_index(p, s, addr), object_map);
+
+       return object_map;
+}
+
+static void put_map(unsigned long *map)
+{
+       VM_BUG_ON(map != object_map);
+       lockdep_assert_held(&object_map_lock);
+
+       spin_unlock(&object_map_lock);
 }
 
 static inline unsigned int size_from_object(struct kmem_cache *s)
@@ -1964,7 +1983,7 @@ static void *get_partial(struct kmem_cache *s, gfp_t flags, int node,
        return get_any_partial(s, flags, c);
 }
 
-#ifdef CONFIG_PREEMPT
+#ifdef CONFIG_PREEMPTION
 /*
  * Calculate the next globally unique transaction for disambiguiation
  * during cmpxchg. The transactions start with the cpu number and are then
@@ -2009,7 +2028,7 @@ static inline void note_cmpxchg_failure(const char *n,
 
        pr_info("%s %s: cmpxchg redo ", n, s->name);
 
-#ifdef CONFIG_PREEMPT
+#ifdef CONFIG_PREEMPTION
        if (tid_to_cpu(tid) != tid_to_cpu(actual_tid))
                pr_warn("due to cpu change %d -> %d\n",
                        tid_to_cpu(tid), tid_to_cpu(actual_tid));
@@ -2341,7 +2360,7 @@ static bool has_cpu_slab(int cpu, void *info)
 
 static void flush_all(struct kmem_cache *s)
 {
-       on_each_cpu_cond(has_cpu_slab, flush_cpu_slab, s, 1, GFP_ATOMIC);
+       on_each_cpu_cond(has_cpu_slab, flush_cpu_slab, s, 1);
 }
 
 /*
@@ -2637,7 +2656,7 @@ static void *__slab_alloc(struct kmem_cache *s, gfp_t gfpflags, int node,
        unsigned long flags;
 
        local_irq_save(flags);
-#ifdef CONFIG_PREEMPT
+#ifdef CONFIG_PREEMPTION
        /*
         * We may have been preempted and rescheduled on a different
         * cpu before disabling interrupts. Need to reload cpu area
@@ -2691,13 +2710,13 @@ redo:
         * as we end up on the original cpu again when doing the cmpxchg.
         *
         * We should guarantee that tid and kmem_cache are retrieved on
* the same cpu. It could be different if CONFIG_PREEMPT so we need
         * to check if it is matched or not.
         */
        do {
                tid = this_cpu_read(s->cpu_slab->tid);
                c = raw_cpu_ptr(s->cpu_slab);
-       } while (IS_ENABLED(CONFIG_PREEMPT) &&
+       } while (IS_ENABLED(CONFIG_PREEMPTION) &&
                 unlikely(tid != READ_ONCE(c->tid)));
 
        /*
@@ -2971,7 +2990,7 @@ redo:
        do {
                tid = this_cpu_read(s->cpu_slab->tid);
                c = raw_cpu_ptr(s->cpu_slab);
-       } while (IS_ENABLED(CONFIG_PREEMPT) &&
+       } while (IS_ENABLED(CONFIG_PREEMPTION) &&
                 unlikely(tid != READ_ONCE(c->tid)));
 
        /* Same with comment on barrier() in slab_alloc_node() */
@@ -3675,13 +3694,12 @@ static void list_slab_objects(struct kmem_cache *s, struct page *page,
 #ifdef CONFIG_SLUB_DEBUG
        void *addr = page_address(page);
        void *p;
-       unsigned long *map = bitmap_zalloc(page->objects, GFP_ATOMIC);
-       if (!map)
-               return;
+       unsigned long *map;
+
        slab_err(s, page, text, s->name);
        slab_lock(page);
 
-       get_map(s, page, map);
+       map = get_map(s, page);
        for_each_object(p, s, addr, page->objects) {
 
                if (!test_bit(slab_index(p, s, addr), map)) {
@@ -3689,8 +3707,9 @@ static void list_slab_objects(struct kmem_cache *s, struct page *page,
                        print_tracking(s, p);
                }
        }
+       put_map(map);
+
        slab_unlock(page);
-       bitmap_free(map);
 #endif
 }
 
@@ -4384,19 +4403,19 @@ static int count_total(struct page *page)
 #endif
 
 #ifdef CONFIG_SLUB_DEBUG
-static void validate_slab(struct kmem_cache *s, struct page *page,
-                                               unsigned long *map)
+static void validate_slab(struct kmem_cache *s, struct page *page)
 {
        void *p;
        void *addr = page_address(page);
+       unsigned long *map;
+
+       slab_lock(page);
 
        if (!check_slab(s, page) || !on_freelist(s, page, NULL))
-               return;
+               goto unlock;
 
        /* Now we know that a valid freelist exists */
-       bitmap_zero(map, page->objects);
-
-       get_map(s, page, map);
+       map = get_map(s, page);
        for_each_object(p, s, addr, page->objects) {
                u8 val = test_bit(slab_index(p, s, addr), map) ?
                         SLUB_RED_INACTIVE : SLUB_RED_ACTIVE;
@@ -4404,18 +4423,13 @@ static void validate_slab(struct kmem_cache *s, struct page *page,
                if (!check_object(s, page, p, val))
                        break;
        }
-}
-
-static void validate_slab_slab(struct kmem_cache *s, struct page *page,
-                                               unsigned long *map)
-{
-       slab_lock(page);
-       validate_slab(s, page, map);
+       put_map(map);
+unlock:
        slab_unlock(page);
 }
 
 static int validate_slab_node(struct kmem_cache *s,
-               struct kmem_cache_node *n, unsigned long *map)
+               struct kmem_cache_node *n)
 {
        unsigned long count = 0;
        struct page *page;
@@ -4424,7 +4438,7 @@ static int validate_slab_node(struct kmem_cache *s,
        spin_lock_irqsave(&n->list_lock, flags);
 
        list_for_each_entry(page, &n->partial, slab_list) {
-               validate_slab_slab(s, page, map);
+               validate_slab(s, page);
                count++;
        }
        if (count != n->nr_partial)
@@ -4435,7 +4449,7 @@ static int validate_slab_node(struct kmem_cache *s,
                goto out;
 
        list_for_each_entry(page, &n->full, slab_list) {
-               validate_slab_slab(s, page, map);
+               validate_slab(s, page);
                count++;
        }
        if (count != atomic_long_read(&n->nr_slabs))
@@ -4452,15 +4466,11 @@ static long validate_slab_cache(struct kmem_cache *s)
        int node;
        unsigned long count = 0;
        struct kmem_cache_node *n;
-       unsigned long *map = bitmap_alloc(oo_objects(s->max), GFP_KERNEL);
-
-       if (!map)
-               return -ENOMEM;
 
        flush_all(s);
        for_each_kmem_cache_node(s, node, n)
-               count += validate_slab_node(s, n, map);
-       bitmap_free(map);
+               count += validate_slab_node(s, n);
+
        return count;
 }
 /*
@@ -4590,18 +4600,17 @@ static int add_location(struct loc_track *t, struct kmem_cache *s,
 }
 
 static void process_slab(struct loc_track *t, struct kmem_cache *s,
-               struct page *page, enum track_item alloc,
-               unsigned long *map)
+               struct page *page, enum track_item alloc)
 {
        void *addr = page_address(page);
        void *p;
+       unsigned long *map;
 
-       bitmap_zero(map, page->objects);
-       get_map(s, page, map);
-
+       map = get_map(s, page);
        for_each_object(p, s, addr, page->objects)
                if (!test_bit(slab_index(p, s, addr), map))
                        add_location(t, s, get_track(s, p, alloc));
+       put_map(map);
 }
 
 static int list_locations(struct kmem_cache *s, char *buf,
@@ -4612,11 +4621,9 @@ static int list_locations(struct kmem_cache *s, char *buf,
        struct loc_track t = { 0, 0, NULL };
        int node;
        struct kmem_cache_node *n;
-       unsigned long *map = bitmap_alloc(oo_objects(s->max), GFP_KERNEL);
 
-       if (!map || !alloc_loc_track(&t, PAGE_SIZE / sizeof(struct location),
-                                    GFP_KERNEL)) {
-               bitmap_free(map);
+       if (!alloc_loc_track(&t, PAGE_SIZE / sizeof(struct location),
+                            GFP_KERNEL)) {
                return sprintf(buf, "Out of memory\n");
        }
        /* Push back cpu slabs */
@@ -4631,9 +4638,9 @@ static int list_locations(struct kmem_cache *s, char *buf,
 
                spin_lock_irqsave(&n->list_lock, flags);
                list_for_each_entry(page, &n->partial, slab_list)
-                       process_slab(&t, s, page, alloc, map);
+                       process_slab(&t, s, page, alloc);
                list_for_each_entry(page, &n->full, slab_list)
-                       process_slab(&t, s, page, alloc, map);
+                       process_slab(&t, s, page, alloc);
                spin_unlock_irqrestore(&n->list_lock, flags);
        }
 
@@ -4682,7 +4689,6 @@ static int list_locations(struct kmem_cache *s, char *buf,
        }
 
        free_loc_track(&t);
-       bitmap_free(map);
        if (!t.count)
                len += sprintf(buf, "No data\n");
        return len;
@@ -2691,13 +2710,13 @@ redo:
         *
         * We should guarantee that tid and kmem_cache are retrieved on
-        * the same cpu. It could be different if CONFIG_PREEMPT so we need
+        * the same cpu. It could be different if CONFIG_PREEMPTION so we need
         * to check if it is matched or not.
         */